package setup

import (
	"context"
	"strings"

	"git.perx.ru/perxis/perxis-go/pkg/errors"

	"git.perx.ru/perxis/perxis-go/pkg/clients"
	"go.uber.org/zap"
)

var (
	ErrCheckClients     = errors.New("clients check error")
	ErrInstallClients   = errors.New("failed to install clients")
	ErrUninstallClients = errors.New("failed to uninstall clients")
)

type (
	Client       = Entity[ClientConf, *clients.Client]
	Clients      = EntityList[ClientConf, *clients.Client]
	ClientOption = EntityOption[ClientConf, *clients.Client]
	ClientConf   struct{}
)

func (ClientConf) Init(e *Entity[ClientConf, *clients.Client]) {
	UpdateExistingClient()(e)
	DeleteClientIfRemoveFlag()(e)
}

var (
	OverwriteClient          = Overwrite[ClientConf, *clients.Client]
	KeepExistingClient       = Keep[ClientConf, *clients.Client]
	DeleteClient             = Delete[ClientConf, *clients.Client]
	DeleteClientIfRemoveFlag = DeleteIfRemoveFlag[ClientConf, *clients.Client]
)

/////

// UpdateExistingClient обновляет существующий клиент, если он существует
// и не содержит необходимых данных
func UpdateExistingClient() ClientOption {
	return func(c *Client) {
		c.UpdateFunc = func(s *Setup, exist, client *clients.Client) (*clients.Client, bool) {
			if exist.Name == "" {
				exist.Name = client.Name
			}

			if exist.Description == "" {
				exist.Description = client.Description
			}

			if exist.OAuth == nil {
				exist.OAuth = client.OAuth
			}

			if exist.TLS == nil {
				exist.TLS = client.TLS
			}

			if exist.APIKey == nil {
				exist.APIKey = client.APIKey
			}

			exist.Disabled = client.Disabled
			exist.RoleID = client.RoleID

			return exist, true
		}
	}
}

//
// Client Setup
//

// AddClient добавляет требования к настройке элементов в пространстве
func (s *Setup) AddClient(client *clients.Client, opt ...ClientOption) *Setup {
	s.Config.Clients.Add(client, opt...)
	return s
}

// AddClients добавляет требования к настройке элементов в пространстве
func (s *Setup) AddClients(clients []*clients.Client, opt ...ClientOption) *Setup {
	s.Config.Clients.AddMany(clients, opt...)
	return s
}

// InstallClient устанавливает клиент в пространстве
func (s *Setup) InstallClient(ctx context.Context, c *Client) error {
	client := c.Value(s)
	client.SpaceID = s.SpaceID

	if s.IsForce() {
		_ = s.content.Clients.Delete(ctx, s.SpaceID, client.ID)
		_, err := s.content.Clients.Create(ctx, client)
		return err
	}

	exist, err := s.content.Clients.Get(ctx, s.SpaceID, client.ID)
	if err != nil {
		if !strings.Contains(err.Error(), clients.ErrNotFound.Error()) {
			return err
		}

		_, err = s.content.Clients.Create(ctx, client)
		return err
	}

	if client, upd := c.UpdateFunc(s, exist, client); upd {
		return s.content.Clients.Update(ctx, client)
	}

	return nil
}

// InstallClients устанавливает все клиенты в пространстве
func (s *Setup) InstallClients(ctx context.Context) error {
	if len(s.Clients) == 0 {
		return nil
	}

	s.logger.Debug("Install clients", zap.String("Space ID", s.SpaceID))

	for _, c := range s.Clients {
		err := s.InstallClient(ctx, c)
		if err != nil {
			client := c.Value(s)
			s.logger.Error("Failed to install client", zap.String("Client ID", client.ID), zap.String("Client Name", client.Name), zap.Error(err))
			return errors.WithDetailf(errors.Wrap(err, "failed to install client"), "Возникла ошибка при настройке клиента %s(%s)", client.Name, client.ID)
		}
	}
	return nil
}

// CheckClient проверяет наличие клиента в пространстве
func (s *Setup) CheckClient(ctx context.Context, c *Client) error {
	client := c.Value(s)
	_, err := s.content.Clients.Get(ctx, s.SpaceID, client.ID)
	return err
}

// CheckClients проверяет наличие всех клиентов в пространстве
func (s *Setup) CheckClients(ctx context.Context) (err error) {
	if len(s.Clients) == 0 {
		return nil
	}

	s.logger.Debug("Check clients", zap.String("Space ID", s.SpaceID))

	var errs []error
	for _, c := range s.Clients {
		err := s.CheckClient(ctx, c)
		if err != nil {
			client := c.Value(s)
			errs = append(errs, errors.WithDetailf(err, "Не найден клиент %s(%s)", client.Name, client.ID))
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrCheckClients, errs...)
	}

	return nil
}

// UninstallClient удаляет клиент из пространства
func (s *Setup) UninstallClient(ctx context.Context, c *Client) error {
	client := c.Value(s)
	if c.DeleteFunc(s, client) {
		if err := s.content.Clients.Delete(ctx, s.SpaceID, client.ID); err != nil && !strings.Contains(err.Error(), clients.ErrNotFound.Error()) {
			return err
		}
	}
	return nil
}

// UninstallClients удаляет все клиенты из пространства
func (s *Setup) UninstallClients(ctx context.Context) error {
	if len(s.Clients) == 0 {
		return nil
	}

	s.logger.Debug("Uninstall clients", zap.String("Space ID", s.SpaceID))

	for _, c := range s.Clients {
		if err := s.UninstallClient(ctx, c); err != nil {
			client := c.Value(s)
			s.logger.Error("Failed to uninstall client", zap.String("Client ID", client.ID),
				zap.String("Client Name", client.Name), zap.Error(err))
			return errors.WithDetailf(errors.Wrap(err, "failed to uninstall client"),
				"Возникла ошибка при удалении клиента %s(%s)", client.Name, client.ID)
		}
	}

	return nil
}
