package setup

import (
	"context"
	"strings"

	"go.uber.org/zap"

	"git.perx.ru/perxis/perxis-go/pkg/data"
	"git.perx.ru/perxis/perxis-go/pkg/errors"
	"git.perx.ru/perxis/perxis-go/pkg/permission"
	"git.perx.ru/perxis/perxis-go/pkg/roles"
)

var (
	ErrCheckRoles     = errors.New("role check error")
	ErrInstallRoles   = errors.New("failed to install role")
	ErrUninstallRoles = errors.New("failed to uninstall role")
)

type Role = Entity[RoleConf, *roles.Role]
type Roles = EntityList[RoleConf, *roles.Role]
type RoleOption = EntityOption[RoleConf, *roles.Role]
type RoleConf struct{}

// Init инициализирует конфигурацию коллекции
func (RoleConf) Init(e *Entity[RoleConf, *roles.Role]) {
	DefaultRoleUpdateFunc()(e)
	DeleteRoleIfRemoveFlag()(e)
}

var (
	NewRole                = NewEntity[RoleConf, *roles.Role]
	OverwriteRole          = Overwrite[RoleConf, *roles.Role]
	KeepExistingRole       = Keep[RoleConf, *roles.Role]
	DeleteRole             = Delete[RoleConf, *roles.Role]
	DeleteRoleIfRemoveFlag = DeleteIfRemoveFlag[RoleConf, *roles.Role]
)

func DefaultRoleUpdateFunc() RoleOption {
	return func(c *Role) {
		c.UpdateFunc = func(s *Setup, exist, new *roles.Role) (*roles.Role, bool) {
			if s.IsForce() {
				return new, true
			}

			if exist.Description == "" {
				exist.Description = new.Description
			}

			if len(exist.Environments) == 0 {
				exist.Environments = new.Environments
			}

			if !data.Contains(s.EnvironmentID, exist.Environments) {
				exist.Environments = append(exist.Environments, s.EnvironmentID)
			}

			exist.Rules = permission.MergeRules(exist.Rules, new.Rules)

			if !exist.AllowManagement {
				exist.AllowManagement = new.AllowManagement
			}

			return exist, true
		}
	}
}

//
// Role Setup
//

// AddRoles добавляет требования к настройке ролей в пространстве
func (s *Setup) AddRoles(roles []*roles.Role, opt ...RoleOption) *Setup {
	s.Config.Roles.AddMany(roles, opt...)
	return s
}

// AddRole добавляет требования к настройке элементов в пространстве
func (s *Setup) AddRole(role *roles.Role, opt ...RoleOption) *Setup {
	s.Config.Roles.Add(role, opt...)
	return s
}

// InstallRole устанавливает роль
func (s *Setup) InstallRole(ctx context.Context, c *Role) error {
	role := c.Value(s)
	role.SpaceID = s.SpaceID

	if !data.Contains(s.EnvironmentID, role.Environments) {
		role.Environments = append(role.Environments, s.EnvironmentID)
	}

	exist, err := s.content.Roles.Get(ctx, s.SpaceID, role.ID)

	// Если роль не найдена, создаем новую
	if err != nil {
		if !strings.Contains(err.Error(), roles.ErrNotFound.Error()) {
			return err
		}
		_, err = s.content.Roles.Create(ctx, role)
		return err
	}

	// Если роль найдена, обновляем ее
	if r, needUpdate := c.UpdateFunc(s, exist, role); needUpdate {
		return s.content.Roles.Update(ctx, r)
	}

	return nil
}

// InstallRoles устанавливает все роли
func (s *Setup) InstallRoles(ctx context.Context) error {
	if len(s.Roles) == 0 {
		return nil
	}

	s.logger.Debug("Install roles", zap.String("Space ID", s.SpaceID), zap.Int("Roles", len(s.Roles)))

	for _, c := range s.Roles {
		if err := s.InstallRole(ctx, c); err != nil {
			role := c.Value(s)
			s.logger.Error("Failed to install role", zap.String("Role ID", role.ID), zap.Error(err))
			return errors.Wrap(errors.WithDetailf(err, "Возникла ошибка при настройке роли %s(%s)", role.ID, role.Description), "failed to install role")
		}
	}

	return nil
}

// UninstallRole удаляет роль
func (s *Setup) UninstallRole(ctx context.Context, c *Role) error {
	role := c.Value(s)
	if c.DeleteFunc(s, role) {
		err := s.content.Roles.Delete(ctx, s.SpaceID, role.ID)
		if err != nil && !strings.Contains(err.Error(), roles.ErrNotFound.Error()) {
			return err
		}
	}
	return nil
}

// UninstallRoles удаляет все роли
func (s *Setup) UninstallRoles(ctx context.Context) error {
	if len(s.Roles) == 0 {
		return nil
	}

	s.logger.Debug("Uninstall role", zap.String("Space ID", s.SpaceID), zap.Int("Roles", len(s.Roles)))

	for _, c := range s.Roles {
		if err := s.UninstallRole(ctx, c); err != nil {
			role := c.Value(s)
			s.logger.Error("Failed to uninstall role", zap.String("Role ID", role.ID), zap.Error(err))
			return errors.WithDetailf(errors.Wrap(err, "failed to uninstall role"),
				"Возникла ошибка при удалении роли %s(%s)", role.ID, role.Description)
		}
	}

	return nil
}

// CheckRole проверяет наличие роли
func (s *Setup) CheckRole(ctx context.Context, c *Role) error {
	role := c.Value(s)
	_, err := s.content.Roles.Get(ctx, s.SpaceID, role.ID)
	return err
}

// CheckRoles проверяет наличие всех ролей
func (s *Setup) CheckRoles(ctx context.Context) error {
	if len(s.Roles) == 0 {
		return nil
	}

	s.logger.Debug("Check role", zap.String("Space ID", s.SpaceID))

	var errs []error
	for _, c := range s.Roles {
		if err := s.CheckRole(ctx, c); err != nil {
			role := c.Value(s)
			errs = append(errs, errors.WithDetailf(err, "Не найдена роль %s(%s)", role.ID, role.Description))
		}
	}

	if len(errs) > 0 {
		return errors.WithErrors(ErrCheckRoles, errs...)
	}

	return nil
}
